/*
**    TP 5ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <mesh_io.hpp>

#include <mesh.hpp>

#include <iostream>
#include <fstream>
#include <sstream>


namespace cpe
{
  mesh mesh_io::load_off_file(const std::string& filename)
  {
    mesh m;

    std::vector<v3> v_vertices;
    std::vector<int> v_connectivity;

    std::ifstream fid(filename.c_str());
    if(!fid.good())
      throw exception_mesh("Cannot open file "+filename,__FILE__,__FUNCTION__,__LINE__);

    std::string buffer;

    //find OFF header
    bool find_off=false;
    while(find_off==false)
    {
      if(!fid.good())
        throw exception_mesh("Cannot find OFF header in file "+filename,__FILE__,__FUNCTION__,__LINE__);
      std::getline(fid,buffer);
      if(buffer.find("OFF")!=std::string::npos)
        find_off=true;
    }

    //read number of vertices + triangles
    int N_vertex=0,N_triangle=0;
    std::getline(fid,buffer);
    std::stringstream tokens(buffer);
    tokens>>N_vertex>>N_triangle;


    //read vertices
    for(int k=0;k<N_vertex;++k)
    {
      std::getline(fid,buffer);
      if(buffer.size()>0 && buffer[0]=='#') {--k; continue;}

      v3 p;
      std::stringstream tokens_vertices(buffer);
      tokens_vertices>>p.x()>>p.y()>>p.z();

      v_vertices.push_back(p);
    }

    //read connectivity
    for(int k=0;k<N_triangle;++k)
    {
      std::getline(fid,buffer);
      if(buffer.size()>0 && buffer[0]=='#') {--k; continue;}

      int N_size,u0,u1,u2;
      std::stringstream tokens_connectivity(buffer);
      tokens_connectivity>>N_size>>u0>>u1>>u2;


      if(N_size!=3)
        throw exception_mesh("Cannot read OFF with non triangular faces for file "+filename,__FILE__,__FUNCTION__,__LINE__);

      v_connectivity.push_back(u0);v_connectivity.push_back(u1);v_connectivity.push_back(u2);
    }

    fid.close();

    m.get_connectivity()=v_connectivity;
    m.get_vertices()=v_vertices;
    return m;

  }

  mesh mesh_io::load_obj_file(const std::string& filename)
  {
    mesh m;


    std::ifstream fid(filename.c_str());
    if(!fid.good())
      throw exception_mesh("Cannot open file "+filename,__FILE__,__FUNCTION__,__LINE__);

    std::string buffer;

    //read the whole file
    while(fid.good()==true)
    {
      //read line
      std::getline(fid,buffer);

      if(buffer.size()>0)
      {
        std::stringstream tokens_buffer(buffer);
        std::string first_word;
        tokens_buffer>>first_word;

        //vertices
        if(first_word=="v")
        {
          v3 vertex;
          tokens_buffer>>vertex.x();
          tokens_buffer>>vertex.y();
          tokens_buffer>>vertex.z();

          m.add_vertex(vertex);
        }

        //texture
        if(first_word=="vt")
        {
          v2 texture;
          tokens_buffer>>texture.x();
          tokens_buffer>>texture.y();

          m.add_texture(texture);
        }
        // normal
        if(first_word=="vn")
        {
          v3 normal;
          tokens_buffer>>normal.x();
          tokens_buffer>>normal.y();
          tokens_buffer>>normal.z();

          m.add_normal(normal);
        }

        // skinning
        if(first_word=="sk")
        { // FIXME! non-standard extension. The Wavefront obj standard allows for extensions, not sure if this one can be used.
          mesh::Skinning skinning;
          tokens_buffer>>skinning.bones[0];
          tokens_buffer>>skinning.weights[0];
          tokens_buffer>>skinning.bones[1];
          tokens_buffer>>skinning.weights[1];
          tokens_buffer>>skinning.bones[2];
          tokens_buffer>>skinning.weights[2];
          tokens_buffer>>skinning.bones[3];
          tokens_buffer>>skinning.weights[3];
          tokens_buffer>>skinning.bones[4];
          tokens_buffer>>skinning.weights[4];
          tokens_buffer>>skinning.bones[5];
          tokens_buffer>>skinning.weights[5];

          m.add_skinning(skinning);
        }

        //connectivity
        if(first_word=="f")
        { // FIXME! We only support triangular faces! And the index for each part (vertex/normal/texture) must be the same (we could assert this)
          std::string i0, i1, i2;
          std::stringstream index;
          int u0=0,u1=0,u2=0;

          tokens_buffer>>i0;
          index.str(i0);
          index>> u0;
          tokens_buffer>>i1;
          index.str(i1);
          index>> u1;
          tokens_buffer>>i2;
          index.str(i2);
          index>> u2;

          m.add_triangle(u0-1,u1-1,u2-1);

        }

      }

    }

    return m;
  }
}
